from __future__ import print_function

import sys

__all__ = ("FileMerge", "FileWindow", "MakeTIFF", "TIFF", "ExifWrapper",
           "identify_raw", "make_pdirs", "raw_exts", "RawWrapper",
           "CommentWrapper", "DotDict", "X3F",)

if sys.version_info[0] > 2:
	basestring = (bytes, str)
	unicode = str
	long = int

class TIFF:
	"""Pretty minimal TIFF container parser"""

	types = { 1: (1, "B"),  # BYTE
	          2: (1, None), # ASCII
	          3: (2, "H"),  # SHORT
	          4: (4, "I"),  # LONG
	          5: (8, "II"), # RATIONAL
	          6: (1, "b"),  # SBYTE
	          7: (1, None), # UNDEFINE
	          8: (2, "h"),  # SSHORT
	          9: (4, "i"),  # SLONG
	         10: (8, "ii"), # SRATIONAL
	         11: (4, "f"),  # FLOAT
	         12: (8, "d"),  # DOUBLE
	         13: (4, "I"),  # IFD
	        }

	def __init__(self, fh, allow_variants=True, short_header=False):
		from struct import unpack
		self._fh = fh
		d = fh.read(4)
		if short_header:
			if d[:2] not in (b"II", b"MM"): raise Exception("Not TIFF")
			self.variant = None
		else:
			good = [b"II*\0", b"MM\0*"]
			if allow_variants:
				# Olympus ORF, Panasonic RW2
				good += [b"IIRO", b"IIU\0"]
			if d not in good: raise Exception("Not TIFF")
			self.variant = d[2:4].strip(b"\0")
		endian = {b"M": ">", b"I": "<"}[d[:1]]
		self._up = lambda fmt, *a: unpack(endian + fmt, *a)
		self._up1 = lambda *a: self._up(*a)[0]
		if short_header:
			next_ifd = short_header
		else:
			next_ifd = self._up1("I", fh.read(4))
		# Be conservative with possibly mis-detected ORF
		if self.variant == b"RO":
			assert next_ifd == 8
		self.reinit_from(next_ifd, short_header)

	def reinit_from(self, next_ifd, short_header=False):
		self.ifd = []
		self.subifd = []
		seen_ifd = set()
		while next_ifd:
			self.ifd.append(self._ifdread(next_ifd))
			if short_header: return
			next_ifd = self._up1("I", self._fh.read(4))
			if next_ifd in seen_ifd:
				from sys import stderr
				print("WARNING: Looping IFDs", file=stderr)
				break
			seen_ifd.add(next_ifd)
			assert len(self.ifd) < 32 # way too many
		subifd = self.ifdget(self.ifd[0], 0x14a) or []
		assert len(subifd) < 32 # way too many
		for next_ifd in subifd:
			self.subifd.append(self._ifdread(next_ifd))

	def ifdget(self, ifd, tag):
		if tag in ifd:
			type, vc, off = ifd[tag]
			if type not in self.types: return None
			if isinstance(off, int): # offset
				self._fh.seek(off)
				tl, fmt = self.types[type]
				off = self._fh.read(tl * vc)
				if fmt: off = self._up(fmt * vc, off)
			if type == 2:
				from ._util import _uni
				off = _uni(off.rstrip(b"\0"))
			return off

	def _ifdread(self, next_ifd):
		ifd = {}
		self._fh.seek(next_ifd)
		count = self._up1("H", self._fh.read(2))
		for i in range(count):
			d = self._fh.read(12)
			tag, type, vc = self._up("HHI", d[:8])
			if type in self.types and self.types[type][0] * vc <= 4:
				tl, fmt = self.types[type]
				d = d[8:8 + (tl * vc)]
				if fmt:
					off = self._up(fmt * vc, d)
				else:
					off = d # ASCII
			else:
				off = self._up1("I", d[8:])
			ifd[tag] = (type, vc, off)
		return ifd

class FileWindow:
	"""A read only view of a range of an fh. You should not continue to use fh."""

	def __init__(self, fh, start=None, length=None):
		if start is None:
			start = fh.tell()
		if length is None or length < 0:
			fh.seek(0, 2)
			stop = fh.tell()
		else:
			stop = start + length
		self.fh = fh
		self.start = start
		self.stop = stop
		self.closed = False
		fh.seek(start)
		assert fh.tell() == start

	def read(self, size=-1):
		if size < 0: size = self.stop - self.fh.tell()
		if size <= 0: return ""
		return self.fh.read(size)

	def tell(self):
		return self.fh.tell() - self.start

	def seek(self, pos, whence=0):
		if whence == 0:
			pos += self.start
		elif whence == 1:
			pos += self.fh.tell()
		else:
			pos += self.stop
		pos = max(min(pos, self.stop), self.start)
		self.fh.seek(pos)

	def close(self):
		if not self.closed:
			self.fh.close()
			self.closed = True

def _parse_date(d):
	from time import strptime, mktime
	def parse(d):
		for fmt in ("%Y:%m:%d %H:%M:%S", "%Y-%m-%d %H:%M:%S", "%Y:%m:%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S"):
			while len(fmt) > 6:
				try:
					return strptime(d, fmt)
				except ValueError:
					pass
				fmt = fmt[:-3]
	tt = parse(d)
	return int(mktime(tt))


class ExifWrapper:
	"""Wrapper for several EXIF libraries.
	Starts out with an internal parser, falls back to GExiv2 for fields
	(and formats) it doesn't know about. (Only works if given a filename.)

	Never fails, just returns empty data (even if file doesn't exist)."""

	def __init__(self, fn):
		self._d = {}
		self._gexiv2_loaded = None
		if isinstance(fn, str):
			self.fn = fn
			fh = None
		else:
			self.fn = None
			fh = fn
		try:
			self._internal(fh)
		except Exception:
			pass

	def _getitem(self, name): # usually overridden from gexiv2
		raise KeyError(name)
	def _contains(self, name): # usually overridden from gexiv2
		return False

	def __getitem(self, name):
		if name in self._d: return self._d[name]
		self._gexiv2()
		v = self._getitem(name)
		if name in ("Exif.Photo.FNumber", "Exif.Photo.FocalLength"):
			a, b = map(int, v.split("/"))
			v = (a, b)
		elif name in (
			"Exif.Image.Orientation", "Exif.Photo.ISOSpeedRatings",
			"Exif.Photo.FocalLengthIn35mmFilm", "Exif.GPSInfo.GPSAltitudeRef",
		):
			v = int(v)
		elif name in ("Exif.GPSInfo.GPSLatitude", "Exif.GPSInfo.GPSLongitude"):
			from fractions import Fraction
			v = [Fraction(*map(int, vv.split("/"))) for vv in v.split()]
		elif name == "Exif.GPSInfo.GPSAltitude":
			a, b = map(float, v.split("/"))
			v = a / b
		return v
	def _fmtrational(self, n, d, intasint=True):
		if n == d == 0: return "0"
		if d == 0: return None # not valid
		if d < 0:
			n = -n
			d = -d
		if d == 1: return str(n)
		if intasint and not n % d: return str(n // d)
		def dec(n, d):
			pos = 0
			while d > 1:
				d //= 10
				pos -= 1
			v = str(n)
			return v[:pos] + "." + v[pos:]
		def rat():
			return "%d/%d" % (n, d)
		def isdec():
			return str(d).rstrip("0") == "1" # A nice decimal number
		if d > n: # < 1, probably a shutter speed, leave it.
			return rat()
		if isdec():
			return dec(n, d)
		if not n % d: return str(n // d) + ".0"
		gcd, tmp = n, d
		while tmp:
			gcd, tmp = tmp, gcd % tmp
		n //= gcd
		d //= gcd
		if isdec():
			return dec(n, d)
		fix = {2: 5, 4: 25, 5: 2, 8: 125, 16: 625, 20: 5, 25: 4, 32: 3125,
		       40: 25, 50: 2, 80: 125, 125: 8, 160: 625, 200: 5, 250: 4,
		       400: 25, 500: 2, 625: 16, 800: 125, 1250: 8, 2000: 5,
		       2500: 4, 3125: 32, 4000: 25, 5000: 2, 6250: 16, 12500: 8,
		       20000: 5, 25000: 4, 50000: 2}
		if d not in fix: return rat()
		return dec(n * fix[d], d * fix[d]).rstrip("0.")
	def __getitem__(self, name):
		v = self.__getitem(name)
		if isinstance(v, tuple):
			if len(v) > 2:
				res = []
				for r in zip(v[::2], v[1::2]):
					r = self._fmtrational(*r)
					if r is None:
						r = "-"
					res.append(r)
				return " ".join(res)
			else:
				return self._fmtrational(*v)
		if hasattr(v, "numerator") and not isinstance(v, (int, long)):
			return self._fmtrational(v.numerator, v.denominator)
		return v

	def __contains__(self, name):
		if name in self._d:
			return True
		self._gexiv2()
		return self._contains(name)

	def _gexiv2(self):
		if self._gexiv2_loaded is not None or not self.fn:
			return
		try:
			import gi
			gi.require_version('GExiv2', '0.10')
			from gi.repository import GExiv2
		except Exception as e:
			self._gexiv2_loaded = False
			print("Failed to load GExiv2:", str(e), file=sys.stderr)
			print("(Maybe you need to install gir1.2-gexiv2-0.10 or similar?)", file=sys.stderr)
			return
		try:
			exif = GExiv2.Metadata(self.fn)
			self._getitem = exif.__getitem__
			self._contains = exif.__contains__
			self._gexiv2_loaded = True
		except Exception:
			self._gexiv2_loaded = False

	def _internal(self, fh=None):
		fh = fh or open(self.fn, "rb")
		try:
			data = fh.read(12)
			if data.startswith(b"FOVb"): # X3F
				fh.seek(0)
				x3f = X3F(fh)
				if x3f.jpegs:
					j = x3f.jpegs[-1]
					fh = FileWindow(fh, j.offset, j.length)
					data = fh.read(12)
				x3f2exif = dict(
					FLENGTH="Exif.Photo.FocalLength",
					FLEQ35MM=("Exif.Photo.FocalLengthIn35mmFilm", lambda v: int(round(float(v))),),
					ISO=("Exif.Photo.ISOSpeedRatings", int,),
					SH_DESC="Exif.Photo.ExposureTime",
					AP_DESC="Exif.Photo.FNumber",
					CAMMODEL="Exif.Image.Model",
					CAMMANUF="Exif.Image.Make",
				)
				for key in set(x3f2exif) & set(x3f.prop):
					value = x3f.prop[key]
					conv = x3f2exif[key]
					if isinstance(conv, tuple):
						value = conv[1](value)
						conv = conv[0]
					self._d[conv] = value
				self._d.update(("X3F." + k, v) for k, v in x3f.prop.items())
			elif data.startswith(b"FUJI"): # RAF
				try:
					offset, length = _RAF_jpeg(fh)
					fh.seek(offset)
					data = fh.read(12)
				except Exception:
					pass
			if data[:3] == b"\xff\xd8\xff": # JPEG
				from struct import unpack
				data = data[3:]
				while data and data[3:7] != b"Exif":
					l = unpack(">H", data[1:3])[0]
					fh.seek(l - 7, 1)
					data = fh.read(9)
					if data[0] == b"\xDA": # Start of Scan
						return
				if not data: return
				# Now comes a complete TIFF, with offsets relative to its' start
				l = unpack(">H", data[1:3])[0]
				fh = FileWindow(fh, length=l)
			# hopefully mostly TIFF (now)
			fh.seek(0)
			tiff = TIFF(fh) # This is the outer TIFF
			ifd0 = tiff.ifd[0]
			exif = tiff.ifdget(ifd0, 0x8769)[0]
			tiff.reinit_from(exif) # replace with Exif IFD(s)
			ifd0.update(tiff.ifd[0]) # merge back into original ifd0
			self._tiff = tiff
			self._ifd = ifd0
			for tag, name, t in ((0x010f, "Exif.Image.Make", False),
			                     (0x0110, "Exif.Image.Model", False),
			                     (0x0112, "Exif.Image.Orientation", False),
			                     (0x829a, "Exif.Photo.ExposureTime", True),
			                     (0x829d, "Exif.Photo.FNumber", True),
			                     (0x8827, "Exif.Photo.ISOSpeedRatings", True),
			                     (0x9003, "Exif.Photo.DateTimeOriginal", False),
			                     (0x9004, "Exif.Photo.CreateDate", False),
			                     (0x920a, "Exif.Photo.FocalLength", True),
			                     (0xa405, "Exif.Photo.FocalLengthIn35mmFilm", False),
			                     (0xa430, "Exif.Photo.OwnerName", False),
			                     (0xa431, "Exif.Photo.SerialNumber", False),
			                     (0xa432, "Exif.Photo.LensInfo", True),
			                     (0xa433, "Exif.Photo.LensMake", False),
			                     (0xa434, "Exif.Photo.LensModel", False),
			                     (0xa435, "Exif.Photo.LensSerialNumber", False),
			                    ):
				val = self._get(tag, t)
				if val is not None: self._d[name] = val
			try:
				self._parse_makernotes()
			except Exception:
				pass
			fl = "Exif.Photo.FocalLength"
			fl135 = "Exif.Photo.FocalLengthIn35mmFilm"
			if (fl135 not in self._d or not self._d[fl135]) and fl in self._d:
				model = self._d.get("Exif.Image.Model")
				n, d = self._d[fl]
				if model.startswith("Canon EOS 5D"):
					self._d[fl135] = int(n / float(d))
				elif model.startswith("E-P") or model.startswith("E-M") or model.startswith("DMC-G"):
					self._d[fl135] = int(n * 2.0 / d)
				elif model == 'Canon PowerShot A590 IS':
					self._d[fl135] = int(n * 5.9 / d)
				elif model == 'Canon PowerShot A610':
					self._d[fl135] = int(n * 4.7 / d)
			gps = self._get(0x8825)
			if gps:
				tiff.reinit_from(gps)
				self._ifd = tiff.ifd[0]
				for tag, name, t in ((0x0001, "Exif.GPSInfo.GPSLatitudeRef", False),
				                     (0x0002, "Exif.GPSInfo.GPSLatitude", True),
				                     (0x0003, "Exif.GPSInfo.GPSLongitudeRef", False),
				                     (0x0004, "Exif.GPSInfo.GPSLongitude", True),
				                     (0x0005, "Exif.GPSInfo.GPSAltitudeRef", False),
				                     (0x0006, "Exif.GPSInfo.GPSAltitude", False),
				                    ):
					val = self._get(tag, t)
					if val is not None:
						from fractions import Fraction
						if t:
							val = [Fraction(*v) for v in  zip(val[::2], val[1::2])]
						self._d[name] = val
		finally:
			fh.close()

	def _get(self, tag, tuple_ok=False):
		d = self._tiff.ifdget(self._ifd, tag)
		if type(d) is tuple:
			if len(d) == 1: return d[0]
			if not tuple_ok: return None
			if d == (): return None
		return d

	def _parse_makernotes(self):
		fh = self._tiff._fh
		if 0x927c in self._ifd:
			type, vc, off = self._ifd[0x927c]
			if type != 7: return
		elif 0xc634 in self._ifd:
			type, vc, off = self._ifd[0xc634]
			if type != 1: return
		elif self.fn: # This is not a wrapped embedded jpeg
			# No makernotes found - try again from embedded jpeg
			# (At least Panasonic RW2 needs this.)
			fh.seek(0)
			j_fh = RawWrapper(fh)._fh
			if j_fh == fh: return
			inner = ExifWrapper(j_fh)
			for k in set(inner._d) - set(self._d):
				self._d[k] = inner._d[k]
			return
		else:
			return
		fh.seek(off)
		data = fh.read(vc)
		if b"\0" not in data: return
		mid = data[:data.find(b"\0")]
		{
			b"AOC"      : self._pentax_makernotes,
			b"PENTAX "  : self._pentax_makernotes,
			b"OLYMPUS"  : self._olympus_makernotes,
			b"Panasonic": self._panasonic_makernotes,
		}.get(mid, lambda _: _)(data, off)
		if self._d.get("Exif.Image.Make") == "Canon" and data[1:4] == "\x00\x01\x00":
			self._canon_makernotes(off)

	def _panasonic_makernotes(self, data, offset):
		# Headerless IFD after "Panasonic\0\0\0"
		# Offsets relative to EXIF block
		if not data.startswith("Panasonic\0\0\0"):
			return
		tiff = self._tiff
		b_ifd, b_subifd = tiff.ifd, tiff.subifd
		try:
			tiff.reinit_from(offset + 12, True)
			lens = tiff.ifdget(tiff.ifd[0], 0x0051)
			if lens:
				self._d["Exif.Panasonic.LensType"] = lens.strip()
		finally:
			tiff.ifd, tiff.subifd = b_ifd, b_subifd

	def _canon_makernotes(self, off):
		# stupid Canon, no header, and offsets relative to EXIF block.
		tiff = self._tiff
		b_ifd, b_subifd = tiff.ifd, tiff.subifd
		try:
			tiff.reinit_from(off)
			lens = tiff.ifdget(tiff.ifd[0], 0x0095)
			if lens:
				self._d["Exif.Canon.LensModel"] = lens
		finally:
			tiff.ifd, tiff.subifd = b_ifd, b_subifd

	def _olympus_makernotes(self, data, offset):
		from io import BytesIO
		fh = BytesIO(data)
		try:
			fh.seek(8)
			t = TIFF(fh, short_header=12)
			offset = t.ifdget(t.ifd[0], 0x2010)[0]
			t.reinit_from(offset)
			data = t.ifdget(t.ifd[0], 0x0203)
			assert isinstance(data, bytes)
			self._d["Exif.OlympusEq.LensModel"] = data.strip()
		except Exception:
			pass

	def _pentax_makernotes(self, data, offset):
		from io import BytesIO
		fh = BytesIO(data[data.find(b"\0") + 1:])
		try:
			t = TIFF(fh, short_header=2)
			lens = " ".join(map(str, t.ifdget(t.ifd[0], 0x3f)))
			self._d["Exif.Pentax.LensType"] = lens
		except Exception:
			pass

	def date(self, tz=None):
		"""Return some reasonable EXIF date field as VTdatetime, or None
		Will guess the timezone (based on environment) if not specified.
		"""
		from wellpapp.vt import VTdatetime
		from time import gmtime, localtime, mktime, strftime
		fields = ("Exif.Photo.DateTimeOriginal", "Exif.Photo.CreateDate", "Exif.Image.DateTime")
		for f in fields:
			try:
				if isinstance(self[f], basestring):
					dt = self[f].strip()
					d, t = dt.split(" ", 1)
					if " " in t:
						t, ltz = t.split(" ")
						ltz = tz or ltz
					elif tz:
						ltz = tz
					else:
						ut = _parse_date(dt)
						lt = list(localtime(ut))
						lt[8] = 0 # ignore dst
						td = int(mktime(gmtime(ut)) - mktime(tuple(lt))) // 60
						ltz = "%s%02d%02d" % ("+" if td <= 0 else "-", abs(td // 60), abs(td % 60))
					date = d.replace(":", "-") + "T" + t + ltz
				else:
					t = int(self[f].strftime("%s"))
					date = strftime("%Y-%m-%dT%H:%M:%SZ", gmtime(t))
				return VTdatetime(date)
			except Exception:
				pass

	def rotation(self):
		if "Exif.Image.Orientation" not in self: return -1
		o = self["Exif.Image.Orientation"]
		orient = {1: 0, 3: 180, 6: 90, 8: 270}
		if o not in orient: return -1
		return orient[o]

raw_exts = ("dng", "pef", "nef", "cr2", "orf", "rw2", "x3f", "raf",)

def _identify_raw(fh, tiff):
	ifd0 = tiff.ifd[0]
	if 0xc612 in ifd0: return "dng"
	if 0x010f in ifd0:
		type, count, off = ifd0[0x010f]
		if type == 2:
			fh.seek(off)
			make = fh.read(min(count, 10))
			if tiff.variant == "*":
				if make[:7] == b"PENTAX ":
					return "pef"
				if make[:6] == b"NIKON ":
					return "nef"
				if make[:5] == b"Canon":
					return "cr2"
			elif tiff.variant == b"RO" and make[:8] == b"OLYMPUS ":
				return "orf"
			elif tiff.variant == b"U" and make == b"Panasonic\0":
				return "rw2"
def identify_raw(fh):
	"""A lower case file extension (e.g. "dng") or None."""
	fh.seek(0)
	data4 = fh.read(4)
	if data4 == b"FOVb":
		return "x3f"
	if data4 == b"FUJI":
		try:
			offset, length = _RAF_jpeg(fh)
			return "raf"
		except Exception:
			pass
	fh.seek(0)
	return _identify_raw(fh, TIFF(fh))

class FileMerge:
	"""Merge ranges of several files"""
	def __init__(self):
		self.contents = []
		self.pos = 0
		self.size = 0
		self.closed = False

	def add(self, fh, pos=0, z=-1):
		"""Add a file, from pos to pos + z"""
		assert pos >= 0
		if z < 0:
			fh.seek(0, 2)
			z = fh.tell() - pos
		assert z > 0
		self.contents.append((fh, pos, z))
		self.size += z

	def close(self):
		if self.closed: return
		done = set()
		for c in self.contents:
			if id(c[0]) not in done:
				done.add(id(c[0]))
				try:
					c[0].close()
				except Exception:
					pass
		self.closed = True

	def read(self, size=-1):
		if self.closed: raise ValueError("I/O operation on closed file")
		max_size = self.size - self.pos
		if size < 0 or size > max_size: size = max_size
		data = b""
		skip = self.pos
		for fh, pos, z in self.contents:
			if skip > z:
				skip -= z
			else:
				fh.seek(pos + skip)
				rsize = min(size, z - skip)
				r = fh.read(rsize)
				assert len(r) == rsize
				data += r
				size -= rsize
				skip = 0
			if not size: break
		self.pos += len(data)
		return data

	def seek(self, offset, whence=0):
		if self.closed: raise ValueError("I/O operation on closed file")
		if not 0 <= whence <= 2: raise IOError("bad whence: " + str(whence))
		if whence == 0:
			pos = offset
		elif whence == 1:
			pos = self.pos + offset
		elif whence == 2:
			pos = self.size + offset
		if pos > self.size: pos = self.size
		if pos < 0: pos = 0
		self.pos = pos

	def tell(self):
		return self.pos

	# IOError instead of AttributeError
	def readline(self):
		raise IOError("FileMerge does not support readline")

class _ThinExif(TIFF):
	"""Pretty minimal TIFF container parser
	Even more minimal now - just for partial Exif parsing"""

	def reinit_from(self, next_ifd, short_header=False):
		self.ifd = self._ifdread(next_ifd)

	def ifdget(self, tag):
		if tag in self.ifd:
			type, vc, off = self.ifd[tag]
			if type not in self.types: return None
			if isinstance(off, int): # offset
				self._fh.seek(off)
				tl, fmt = self.types[type]
				off = self._fh.read(tl * vc)
				if fmt: off = self._up(fmt * vc, off)
			return type, off

class MakeTIFF:
	types = { 1: "B", # BYTE
	          3: "H", # SHORT
	          4: "I", # LONG
	          5: "I", # RATIONAL
	          6: "b", # SBYTE
	          8: "h", # SSHORT
	          9: "i", # SLONG
	         10: "i", # SRATIONAL
	         11: "f", # FLOAT
	         12: "d", # DOUBLE
	        }

	def __init__(self):
		self.entries = {}

	def add(self, tag, values):
		self.entries[tag] = values

	def _one(self, tag, data):
		from struct import pack
		type, values = data
		if isinstance(values, unicode):
			values = values.encode("utf-8")
		if isinstance(values, bytes):
			d = pack(">HHI", tag, 2, len(values))
		else:
			f = self.types[type]
			z = len(values)
			if type == 5: z //= 2
			values = pack(">" + (f * len(values)), *values)
			d = pack(">HHI", tag, type, z)
			if len(values) <= 4:
				d += (values + b"\x00\x00\x00")[:4]
				values = None
		if values:
			d += pack(">I", self._datapos)
			values = (values + b"\x00\x00\x00")[:((len(values) + 3) // 4) * 4]
			self._data += values
			self._datapos += len(values)
		return d

	def serialize(self, offset=0):
		from struct import pack
		data = pack(">ccHIH", b"M", b"M", 42, 8, len(self.entries))
		self._datapos = 10 + 12 * len(self.entries) + 4 + offset
		self._data = b""
		for tag in sorted(self.entries):
			data += self._one(tag, self.entries[tag])
		data += pack(">I", 0)
		data += self._data
		return data

def _rawexif(raw, fh):
	from struct import pack, unpack
	from io import BytesIO
	fm = FileMerge()
	def read_marker():
		while 42:
			d = fh.read(1)
			if d != b"\xFF": return d
	fh.seek(0)
	assert read_marker() == b"\xD8"
	first = fh.tell()
	marker = read_marker()
	while marker == b"\xE0": # APP0, most likely JFIF
		first = fh.tell() + unpack(">H", fh.read(2))[0] - 2
		fh.seek(first)
		marker = read_marker()
	fm.add(fh, 0, first)
	if marker == b"\xE1": # APP1, most likely Exif
		candidate = fh.tell() + unpack(">H", fh.read(2))[0] - 2
		if fh.read(5) == b"Exif\x00":
			# Already has EXIF, leave as is.
			fh.seek(0)
			return fh
		first = candidate
		fh.seek(first)
		marker = read_marker()
	raw.seek(0)
	exif0 = MakeTIFF()
	exif1 = MakeTIFF()
	raw_exif = _ThinExif(raw)
	for k in 0x010f, 0x0110, 0x0112, 0x0131, 0x0132, 0x013b, 0x8298, 0xc614:
		d = raw_exif.ifdget(k)
		if d: exif0.add(k, d)
	offset = len(exif0.serialize()) + 12
	exif0.add(0x8769, (4, (offset,)))
	exif_ifd = raw_exif.ifd[0x8769][2][0]
	raw_exif.reinit_from(exif_ifd)
	for k in raw_exif.ifd:
		if k == 0x927c: continue # Skip makernotes
		d = raw_exif.ifdget(k)
		if d: exif1.add(k, d)
	exif = b"Exif\x00\x00"
	exif += exif0.serialize() + exif1.serialize(offset - 8)[8:]
	exif = b"\xFF\xE1" + pack(">H", len(exif) + 2) + exif
	fm.add(BytesIO(exif))
	fm.add(fh, first)
	return fm

def _RAF_jpeg(fh):
	from struct import unpack
	fh.seek(84)
	offset, length = unpack(">II", fh.read(8))
	fh.seek(offset)
	if fh.read(3) == "\xff\xd8\xff":
		return offset, length

class X3F:
	"""Find JPEG sections and metadata in X3F files
	.jpegs is a list from smallest to largest.
	.prop is a dict {field: value}
	"""

	def __init__(self, fh):
		from collections import defaultdict, namedtuple
		self.X3F_JPEG = namedtuple("X3F_JPEG", "offset length cols rows")
		from struct import unpack
		assert fh.read(4) == b"FOVb"
		self.vminor, self.vmajor = unpack("<HH", fh.read(4))
		assert self.vmajor == 2
		fh.seek(-4, 2)
		index, = unpack("<I", fh.read(4))
		fh.seek(index)
		assert fh.read(4) == b"SECd" # section header
		secvminor, secvmajor = unpack("<HH", fh.read(4))
		assert secvmajor == 2
		nentries, = unpack("<I", fh.read(4))
		sections = defaultdict(list)
		for i in range(nentries):
			offset, length = unpack("<II", fh.read(8))
			typ = fh.read(4)
			sections[typ].append((offset, length,))
		self.jpegs = []
		for offset, length in sections[b"IMAG"] + sections[b"IMA2"]:
			fh.seek(offset)
			self._read_img(fh, length)
		self.jpegs.sort(key=lambda j: j.cols * j.rows)
		self.prop = {}
		for offset, length in sections[b"PROP"]:
			self._read_prop(fh, offset)

	def _read_img(self, fh, length):
		from struct import unpack
		assert fh.read(4) == b"SECi"
		vminor, vmajor = unpack("<HH", fh.read(4))
		assert vmajor == 2
		typ, = unpack("<I", fh.read(4))
		if typ != 2:
			return
		fmt, cols, rows, rowz = unpack("<IIII", fh.read(16))
		if fmt != 18:
			return
		offset = fh.tell()
		length -= 28
		self.jpegs.append(self.X3F_JPEG(offset=offset, length=length, cols=cols, rows=rows))

	def _read_prop(self, fh, offset):
		from struct import unpack
		fh.seek(offset)
		assert fh.read(4) == b"SECp"
		vminor, vmajor = unpack("<HH", fh.read(4))
		assert vmajor == 2
		nentries, fmt, _, totlen = unpack("<IIII", fh.read(16))
		if fmt != 0:
			return
		hdrs = fh.read(nentries * 8)
		offset = fh.tell()
		# Stupid format
		def getstr(str_off):
			fh.seek(offset + str_off*2)
			def reader():
				while True:
					v = fh.read(2)
					if v == b"\0\0" or not v:
						raise StopIteration
					yield v
			return b"".join(reader()).decode("utf-16le")
		while hdrs:
			name_off, value_off = unpack("<II", hdrs[:8])
			hdrs = hdrs[8:]
			self.prop[getstr(name_off)] = getstr(value_off)

class RawWrapper:
	"""Wraps (read only) IO to an image, so that RAW images look like JPEGs.
	Handles DNG, NEF, PEF, CR2, ORF, X3F and RAF.
	Wraps fh as is if no reasonable embedded JPEG is found."""

	def __init__(self, fh, make_exif=False):
		self.closed = False
		self._set_fh(fh)
		fh.seek(0)
		data4 = fh.read(4)
		if data4 == b"FOVb":
			try:
				fh.seek(0)
				x3f = X3F(fh)
				if x3f.jpegs:
					j = x3f.jpegs[-1]
					self._test_jpeg([j.offset], [j.length])
					# No make_exif support, but at least Sigma SD14 puts exif in this JPEG.
					return
			except Exception:
				pass
		elif data4 == b"FUJI":
			try:
				offset, length = _RAF_jpeg(fh)
				self._test_jpeg([offset], [length])
				# No make_exif support, but at least X-T2 puts exif in this JPEG.
				return
			except Exception:
				pass
		fh.seek(0)
		try:
			tiff = TIFF(self)
			fmt = _identify_raw(self, tiff)
			if fmt == "dng" and len(tiff.subifd) > 1:
				jpeg = tiff.ifdget(tiff.subifd[1], 0x111)
				jpeglen = tiff.ifdget(tiff.subifd[1], 0x117)
				self._test_jpeg(jpeg, jpeglen)
			elif fmt == "nef" and tiff.subifd:
				jpeg = tiff.ifdget(tiff.subifd[0], 0x201)
				jpeglen = tiff.ifdget(tiff.subifd[0], 0x202)
				self._test_jpeg(jpeg, jpeglen)
			elif fmt == "pef":
				self._test_pef(tiff)
			elif fmt == "cr2":
				self._test_cr2(tiff)
			elif fmt == "orf":
				self._test_orf(tiff)
			elif fmt == "rw2":
				self._test_rw2(tiff)
		except Exception:
			pass
		self.seek(0)
		if make_exif and self._fh != fh:
			self._set_fh(_rawexif(fh, self._fh))

	def _set_fh(self, fh):
		self._fh = fh
		self.read = fh.read
		self.seek = fh.seek
		self.tell = fh.tell

	def close(self):
		if not self.closed:
			self._fh.close()
			self.closed = True

	def _test_pef(self, tiff):
		for ifd in tiff.ifd[1:]:
			w, h = tiff.ifdget(ifd, 0x100), tiff.ifdget(ifd, 0x101)
			if w and h and max(w[0], h[0]) > 1000: # looks like a real image
				jpeg = tiff.ifdget(ifd, 0x201)
				jpeglen = tiff.ifdget(ifd, 0x202)
				if self._test_jpeg(jpeg, jpeglen):
					return True

	def _test_cr2(self, tiff):
		for ifd in tiff.ifd:
			w, h = tiff.ifdget(ifd, 0x100), tiff.ifdget(ifd, 0x101)
			if w and h and max(w[0], h[0]) > 1000: # looks like a real image
				jpeg = tiff.ifdget(ifd, 0x111)
				jpeglen = tiff.ifdget(ifd, 0x117)
				if self._test_jpeg(jpeg, jpeglen):
					return True

	def _test_orf(self, tiff):
		try:
			exif = tiff.ifdget(tiff.ifd[0], 0x8769)[0]
			tiff.reinit_from(exif)
			makernotes = tiff.ifd[0][0x927c][2]
			tiff._fh.seek(makernotes)
			assert tiff._fh.read(10) == b"OLYMPUS\0II"
			tiff.reinit_from(makernotes + 12)
			jpegifd = tiff.ifdget(tiff.ifd[0], 0x2020)[0]
			tiff.reinit_from(makernotes + jpegifd)
			jpegpos = tiff.ifdget(tiff.ifd[0], 0x101)[0]
			jpeglen = tiff.ifdget(tiff.ifd[0], 0x102)[0]
			return self._test_jpeg([makernotes + jpegpos], [jpeglen])
		except Exception:
			pass

	def _test_rw2(self, tiff):
		try:
			type, vc, off = tiff.ifd[0][0x2e]
			if type == 7:
				return self._test_jpeg([off], [vc])
		except Exception:
			pass

	def _test_jpeg(self, jpeg, jpeglen):
		if not jpeg or not jpeglen: return
		if len(jpeg) != len(jpeglen): return
		jpeg, jpeglen = jpeg[-1], jpeglen[-1]
		self.seek(jpeg)
		if self.read(3) == b"\xff\xd8\xff":
			self._set_fh(FileWindow(self._fh, jpeg, jpeglen))
			return True

	# This is needed so PIL gives an IOError and not an AttributeError on some files.
	def readline(self):
		raise IOError("RawWrapper doesn't support readline.")

def make_pdirs(fn):
	"""Like mkdir -p `dirname fn`"""

	import os.path
	dn = os.path.dirname(fn)
	if not os.path.exists(dn): os.makedirs(dn)

class CommentWrapper:
	"""Wrap a file so readline/iteration skips comments
	and optionally empty lines"""

	def __init__(self, fh, allow_empty=False):
		self.fh = fh
		self.allow_empty = allow_empty
		self.close = fh.close

	def __iter__(self):
		return self

	def __next__(self):
		line = self.readline()
		if not line: raise StopIteration()
		return line
	next = __next__

	def readline(self):
		while 42:
			line = self.fh.readline()
			if not line: return line
			s = line.strip()
			if s:
				if s[0] != "#": return line
			elif self.allow_empty:
				return line

	def __enter__(self):
		return self

	def __exit__(self, *exc):
		self.close()

class DotDict(dict):
	"""Like a dict, but with d.foo as well as d["foo"]."""

	__setattr__ = dict.__setitem__
	__delattr__ = dict.__delitem__
	def __getattr__(self, name):
		if name[0] == "_":
			raise AttributeError(name)
		return self.get(name)
	def __repr__(self):
		return repr(type(self)) + dict.__repr__(self)




from __future__ import print_function

from os.path import lexists, realpath
from os import unlink, rename, symlink, stat
from optparse import OptionParser
from hashlib import md5
from PIL import Image, PngImagePlugin
from wellpapp import Client, make_pdirs

def main(arg0, argv):
	p = OptionParser(usage="Usage: %prog [-t] post-spec new-file", prog=arg0)
	p.add_option("-t", "--regenerate-thumbs",
	             action="store_true",
	             help="Regenerate thumbnails from new-file"
	            )
	opts, args = p.parse_args(argv)
	if len(args) != 2:
		p.print_help()
		return 1

	client = Client()

	oldfile, newfile = args
	m = client.postspec2md5(oldfile)
	post = client.get_post(m, wanted=(["ext", "rotate"]))
	if not post:
		print("Post not found")
		return 1
	data = open(newfile, "rb").read()
	newm = md5(data).hexdigest()
	mtime = stat(newfile).st_mtime
	if client.get_post(newm, wanted=()):
		print("New file already has post")
		return 1
	path = client.image_path(newm)
	if lexists(path):
		unlink(path)
	make_pdirs(path)
	symlink(realpath(newfile), path)
	if opts.regenerate_thumbs:
#		@@ assumes same ext
		client.save_thumbs(newm, None, post.ext, post.rotate, True)
	else:
		meta = PngImagePlugin.PngInfo()
		meta.add_text("Thumb::URI", str(newm + "." + post.ext), 0)
		meta.add_text("Thumb::MTime", str(int(mtime)), 0)
	sizes = list(map(int, client.cfg.thumb_sizes.split())) + ["normal", "large"]
	for z in sizes:
		if isinstance(z, int):
			oldpath = client.thumb_path(m, z)
			if opts.regenerate_thumbs:
				unlink(oldpath)
			else:
				newpath = client.thumb_path(newm, z)
				make_pdirs(newpath)
				rename(oldpath, newpath)
		else:
			oldpath = client.pngthumb_path(m, post.ext, z)
			if opts.regenerate_thumbs:
				unlink(oldpath)
			else:
				t = Image.open(oldpath)
				t.load()
				newpath = client.pngthumb_path(newm, post.ext, z)
				make_pdirs(newpath)
				t.save(newpath, format="PNG", pnginfo=meta)
	client.modify_post(m, MD5=newm)
	path = client.image_path(m)
	if lexists(path):
		unlink(path)



from __future__ import print_function

import sys
from importlib import import_module

def main():
	arg0 = sys.argv[0]
	argv = sys.argv[1:]
	def usage(fh):
		print("Usage:", arg0, "command [args]", file=fh)
	if len(argv) < 1:
		usage(sys.stderr)
		return 1
	command = argv[0]
	argv = argv[1:]
	if command in ('-h', '--help'):
		usage(sys.stdout)
		return 0
	modname = 'wellpapp.shell.' + command
	if command.startswith('_') or '.' in command:
		modname = 'wellpapp._' # anything that won't be found
	try:
		mod = import_module(modname)
	except ModuleNotFoundError as e:
		if e.name != modname:
			raise
		print("Unknown command", command, file=sys.stderr)
		return 1
	return mod.main('%s %s' % (arg0, command), argv)

if __name__ == '__main__':
	sys.exit(main())



# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import division

from sys import version_info
if version_info[0] > 2:
	unicode = str
	unichr = chr
	ord = int
	import queue as Queue
else:
	import Queue

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk
from gi.repository import GdkPixbuf
from gi.repository.GObject import TYPE_STRING, TYPE_INT
from gi.repository import Pango
from gi.repository.GLib import markup_escape_text, idle_add

from itertools import chain
from os.path import commonprefix
from hashlib import md5
from threading import Thread
from multiprocessing import cpu_count

from wellpapp import Client, RawWrapper, WellpappError

def clean(n):
	n = _uni(n)
	if n[0] in u"-~": return n[1:]
	return n
def prefix(n):
	n = _uni(n)
	if n[0] in u"-~": return n[0]
	return u""

def ishex(s):
	return all(c in "1234567890abcdef" for c in s.lower())

def _uni(s):
	if not isinstance(s, unicode):
		try:
			s = s.decode("utf-8")
		except UnicodeDecodeError:
			s = s.decode("iso-8859-1")
	return s

_fuzz_ignore = u"".join(map(unichr, range(33))) + u"-_()[]{}.,!/\"'?<>@=+%$#|\\"
def _completefuzz(word):
	return [c for c in word.lower() if c not in _fuzz_ignore]

def complete(tw, word):
	assert u" " not in word
	pre = prefix(word)
	word = clean(word)
	fuzz_word = _completefuzz(word)
	if pre == "-":
		known_tags = set()
		if tw.thumbview.get_selected_items():
			lists = ("all", "allcurrent", "currentother")
		else:
			lists = ("any",)
		for li in lists:
			known_tags.update(clean(t) for t in tw.taglist[li])
		known_tags = [tw.ids[t] for t in known_tags]
		def gen():
			yield [t for t in known_tags if t.name.startswith(word)]
			yield [t for t in known_tags if any(a.startswith(word) for a in t.get("alias", ()))]
			yield [t for t in known_tags if _completefuzz(t.name)[:len(fuzz_word)] == fuzz_word]
			yield [t for t in known_tags if any(_completefuzz(a)[:len(fuzz_word)] == fuzz_word for a in t.get("alias", ()))]
		tags_gen = gen()
	else:
		tags_gen = (tw.client.find_tags(t, word) for t in ("EI", "EAI", "FI", "FAI"))
	for tags in tags_gen:
		if len(tags) == 1:
			candidates = [tags[0].name] + tags[0].get("alias", [])
			for name in candidates:
				if name.startswith(word):
					break
			else:
				for name in candidates:
					if _completefuzz(name)[:len(fuzz_word)] == fuzz_word:
						break
				else:
					name = candidates[0]
			return pre + name, [(name, tags[0])]
		if len(tags) > 1: break
	names = {}
	for t in tags:
		names[t.name] = t
		if "alias" in t:
			for n in t.alias:
				names[n] = t
	aliases = [t["alias"] if "alias" in t else [] for t in tags]
	aliases = list(chain(*aliases))
	tags = [t["name"] for t in tags]
	inc = lambda n: n[:len(word)] == word
	candidates = list(filter(inc, tags)) + list(filter(inc, aliases))
	if not candidates:
		inc = lambda n: _completefuzz(n)[:len(fuzz_word)] == fuzz_word
		candidates = list(filter(inc, tags)) + list(filter(inc, aliases))
	common = commonprefix(candidates)
	if _completefuzz(common)[:len(fuzz_word)] != fuzz_word:
		common = word
	word = pre + common
	candidates = [(c, names[c]) for c in candidates]
	return word, candidates

def complete_entry(tw, parent, window, tagfield, event):
	if gdk.keyval_name(event.keyval) == "Tab" and \
	   not event.state & (gdk.ModifierType.SHIFT_MASK | gdk.ModifierType.CONTROL_MASK):
		parent.tab_count += 1
		text = _uni(tagfield.get_text())
		pos = tagfield.get_position()
		spos = text.rfind(u" ", 0, pos - 1) + 1
		left = text[:spos]
		word = text[spos:pos].strip()
		right = text[pos:]
		def update(new_word):
			text = left + new_word + right
			tagfield.set_text(text)
			tagfield.set_position(pos + len(new_word) - len(word))
		if parent.tab_count == 1:
			parent.show_alts([])
			if word:
				new_word, alts = complete(tw, word)
				parent.tab_alts = alts
				if alts:
					parent.show_alts(alts)
					if new_word:
						if len(alts) == 1 and (not right or right[0] != u" "):
							new_word += u" "
						update(new_word)
		elif parent.tab_count == 2:
			parent.tab_count = 0
			if parent.tab_alts:
				left = left + prefix(word)
				word = clean(word)
				dialog = TagCompletionDialog(tw, tw.client, window, word, parent.tab_alts)
				if dialog.run() == gtk.ResponseType.ACCEPT:
					update(dialog.get_tt())
				dialog.destroy()
		return True
	parent.tab_count = 0

class FixedTreeView(gtk.TreeView):
	def __init__(self, **kw):
		gtk.TreeView.__init__(self, **kw)
		self._event = None
		self._select_all_state = False
		self.get_selection().set_select_function(self._sel)
		self.connect('button-press-event', self._press)
		self.connect('button-release-event', self._release)
		self.connect('select-all', self._select_all)

	def _sel(self, selection, model, path, current):
		if self._select_all_state:
			return model[path[0]][5] != "impl"
		return self._event == None

	def _press(self, tv, event, data=None):
		self._select_all_state = False
		if event.state & (gdk.ModifierType.CONTROL_MASK | gdk.ModifierType.SHIFT_MASK):
			return
		if event.button == 2:
			return
		ev = (int(event.x), int(event.y))
		path = self.get_path_at_pos(*ev)
		if not path: return True
		if self.get_selection().path_is_selected(path[0]):
			self._event = ev
		else:
			self._event = None

	def _release(self, tv, event, data=None):
		self._select_all_state = False
		if self._event:
			oldev = self._event
			self._event = None
			if oldev != (int(event.x), int(event.y)): return True
			path = self.get_path_at_pos(*oldev)
			if path: self.set_cursor(path[0], path[1])

	def _select_all(self, self_again):
		self._select_all_state = not self._select_all_state

taglisttypes = [TYPE_STRING] * 6

# In GTK3 the mnemonics are really slow, so I'll fake it with manual
# underlining and an AccelGroup.
def button_with_mnemonic(ag, text, func, *args):
	cb = lambda *a: func(*args)
	button = gtk.Button.new_with_label(text)
	apos = text.index(u"_")
	accel = text[apos + 1]
	key, mod = gtk.accelerator_parse("<Alt>" + accel)
	ag.connect(key, mod, 0, cb)
	text = u"%s<u>%s</u>%s" % (text[:apos], accel, text[apos + 2:])
	button.get_child().set_markup(text)
	button.connect("clicked", cb)
	return button

class TagWindow:
	def __init__(self):
		self.client = Client()
		self.window = gtk.Window(type=gtk.WindowType.TOPLEVEL)
		self.window.set_border_width(2)
		self.window.connect("destroy", self.destroy)
		self.bbox = gtk.HBox(homogeneous=False, spacing=0)
		ag = gtk.AccelGroup()
		self.window.add_accel_group(ag)
		self.b_apply = button_with_mnemonic(ag, u"_Apply", self.apply_action)
		self.b_apply.set_sensitive(False)
		self.b_quit = button_with_mnemonic(ag, u"_Quit", self.destroy)
		self.bbox.pack_start(self.b_apply, True, True, 0)
		self.bbox.pack_end(self.b_quit, False, False, 0)
		self.msg = gtk.Label(label=u"Starting..")
		self.msg.set_ellipsize(Pango.EllipsizeMode.END)
		self.msgbox = gtk.EventBox()
		self.msgbox.add(self.msg)
		self.progress_bar = gtk.ProgressBar()
		self.thumbs = gtk.ListStore(TYPE_STRING, GdkPixbuf.Pixbuf, TYPE_STRING)
		self.thumbview = gtk.IconView(model=self.thumbs)
		self.thumbview.set_pixbuf_column(1)
		self.thumbview.set_tooltip_column(2)
		self.thumbview.set_selection_mode(gtk.SelectionMode.MULTIPLE)
		self.thumbview.connect("selection-changed", self.thumb_selected)
		self.thumbview.connect("item-activated", self.thumb_activated)
		self.thumbview.connect("button-press-event", self.middle_toggle_select)
		self.tagbox = gtk.VBox(homogeneous=False, spacing=0)
		self.tags_all = gtk.ListStore(*taglisttypes)
		self.tags_allcurrent = gtk.ListStore(*taglisttypes)
		self.tags_currentother = gtk.ListStore(*taglisttypes)
		self.tags_other = gtk.ListStore(*taglisttypes)
		self.tags_allview = FixedTreeView(model=self.tags_all)
		celltext = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("ALL", celltext, markup=0)
		tvc.add_attribute(celltext, "cell-background", 2)
		tvc.add_attribute(celltext, "foreground", 3)
		self.tags_allview.append_column(tvc)
		self.tags_allcurrentview = FixedTreeView(model=self.tags_allcurrent)
		celltext = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("All Current", celltext, markup=0)
		tvc.add_attribute(celltext, "cell-background", 2)
		tvc.add_attribute(celltext, "foreground", 3)
		self.tags_allcurrentview.append_column(tvc)
		self.tags_currentotherview = FixedTreeView(model=self.tags_currentother)
		celltext = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("Some Current", celltext, markup=0)
		tvc.add_attribute(celltext, "cell-background", 2)
		tvc.add_attribute(celltext, "foreground", 3)
		self.tags_currentotherview.append_column(tvc)
		self.tags_otherview = FixedTreeView(model=self.tags_other)
		for tv in self.tags_allview, self.tags_allcurrentview, \
		          self.tags_currentotherview, self.tags_otherview:
			tv.connect("row-activated", self.modify_tag)
			tv.connect("button-press-event", self.middle_toggle_select)
		celltext = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("Some", celltext, markup=0)
		tvc.add_attribute(celltext, "cell-background", 2)
		tvc.add_attribute(celltext, "foreground", 3)
		self.tags_otherview.append_column(tvc)
		guidtype = gtk.TargetEntry.new("text/x-wellpapp-tagguid", 0, 1)
		nametype = gtk.TargetEntry.new("text/x-wellpapp-tagname", 0, 4)
		posttype = gtk.TargetEntry.new("text/x-wellpapp-post-id", 0, 0)
		texttypes = [gtk.TargetEntry.new("STRING", 0, 4), gtk.TargetEntry.new("text/plain", 0, 4)]
		srctypes = texttypes + [nametype, guidtype]
		for widget in self.tags_allview, self.tags_allcurrentview, self.tags_currentotherview, self.tags_otherview:
			widget.drag_source_set(gdk.ModifierType.BUTTON1_MASK, srctypes, gdk.DragAction.COPY)
			widget.drag_source_set_icon_name("text-x-generic")
			widget.connect("drag_data_get", self.drag_get_list)
		for widget, all in (self.tags_allview, True), (self.tags_allcurrentview, False):
			widget.drag_dest_set(gtk.DestDefaults.ALL, [guidtype], gdk.DragAction.COPY)
			widget.connect("drag_data_received", self.drag_put, all)
		self.thumbview.drag_source_set(gdk.ModifierType.BUTTON1_MASK, [posttype] + texttypes, gdk.DragAction.COPY)
		self.thumbview.connect("drag_data_get", self.drag_get_icon)
		self.thumbview.connect("drag_begin", self.drag_icon_begin)
		self.thumbview.drag_dest_set(gtk.DestDefaults.ALL, [guidtype, posttype], gdk.DragAction.COPY)
		self.thumbview.connect("drag_data_received", self.drag_put_thumb)
		self.tagbox.pack_start(self.tags_allview, False, False, 0)
		self.tagbox.pack_start(self.tags_allcurrentview, False, False, 0)
		self.tagbox.pack_start(self.tags_currentotherview, False, False, 0)
		self.tagbox.pack_start(self.tags_otherview, False, False, 0)
		self.mbox = gtk.HPaned()
		self.mbox.set_position(650)
		self.thumbscroll = gtk.ScrolledWindow()
		self.thumbscroll.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.ALWAYS)
		self.thumbscroll.add(self.thumbview)
		self.tagscroll = gtk.ScrolledWindow()
		self.tagscroll.set_size_request(150, -1)
		self.tagscroll.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
		tagbox_wp = gtk.Viewport()
		tagbox_wp.add(self.tagbox)
		self.tagscroll.add(tagbox_wp)
		self.mbox.pack1(self.thumbscroll, resize=True, shrink=False)
		self.mbox.pack2(self.tagscroll, resize=True, shrink=False)
		self.vbox = gtk.VBox(homogeneous=False, spacing=0)
		self.vbox.pack_start(self.msgbox, False, False, 0)
		self.vbox.pack_start(self.progress_bar, False, False, 0)
		self.vbox.pack_start(self.mbox, True, True, 0)
		self.vbox.pack_end(self.bbox, False, False, 0)
		self.tagfield = gtk.Entry()
		self.tagfield.set_placeholder_text('tag tag tag')
		self.tagfield.connect("activate", self.apply_action, None)
		self.tagfield.connect("key-press-event", self.tagfield_key)
		key, mod = gtk.accelerator_parse('<Alt>s')
		ag.connect(key, mod, 0, self._focus_tagfield)
		self.tagfield.drag_dest_set(gtk.DestDefaults.ALL, [nametype] + texttypes, gdk.DragAction.COPY)
		self.tagfield.connect("drag_data_received", self.drag_put_tagfield)
		self.vbox.pack_end(self.tagfield, False, False, 0)
		self.window.add(self.vbox)
		self.window.set_default_size(int(self.client.cfg.tagwindow_width), int(self.client.cfg.tagwindow_height))
		self.window.show_all()
		# msgbox and progress_bar occupy the same space (one at a time), so make them the same size
		widgets = (self.msgbox, self.progress_bar,)
		y = max(widget.get_preferred_size().natural_size.height for widget in widgets)
		for widget in widgets:
			widget.set_size_request(-1, y)
		self.progress_bar.hide()
		self.type2colour = dict([cs.split("=") for cs in self.client.cfg.tagcolours.split()])
		self.fullscreen_open = False
		self.taglist = {}
		for pre in ("", "impl"):
			for suf in ("any", "all", "allcurrent", "currentother", "other"):
				self.taglist[pre + suf] = set()
		self.tab_count = 0
		self.tab_alts = []
		self._tagfield_prev_ctx = None

	def _focus_tagfield(self, *a):
		self.tagfield.grab_focus()
		self.tagfield.select_region(-1, -1)

	def drag_icon_begin(self, widget, ctx):
		for path in widget.get_selected_items():
			pixbuf = widget.get_model()[path][1]
			widget.drag_source_set_icon_pixbuf(pixbuf)
			break
		else:
			widget.drag_source_set_icon_name("image")

	def drag_put_tagfield(self, widget, context, x, y, selection, targetType, eventTime):
		# This gets called twice (why?), so ignore it the second time
		if self._tagfield_prev_ctx != context:
			self._tagfield_prev_ctx = context
			tag = _uni(selection.get_data()) + u" "
			text = _uni(self.tagfield.get_text())
			if text and text[-1] not in (u" ", u"="): text += u" "
			text += tag
			self.tagfield.set_text(text)
		# When recieving standard types, we also have to stop the default handler
		context.finish(True, False, eventTime)
		widget.emit_stop_by_name("drag_data_received")

	def drag_put_thumb_guid(self, x, y, selection):
		x += int(self.thumbscroll.get_hadjustment().get_value())
		y += int(self.thumbscroll.get_vadjustment().get_value())
		item = self.thumbview.get_item_at_pos(x, y)
		if not item: return
		iter = self.thumbs.get_iter(item[0])
		m = self.thumbs.get_value(iter, 0)
		# @@ check for changing tag values?
		self._apply([((t, None), None) for t in _uni(selection.get_data()).split()], [], [m])

	def drag_put_thumb_post(self, selection):
		try:
			data = _uni(selection.get_data()).lower()
		except Exception:
			from traceback import print_exc
			print_exc()
			return
		md5s = {t[0] for t in self.thumbs}
		add = [m for m in data.split() if len(m) == 32 and ishex(m) and m not in md5s]
		if add:
			fl = FileLoader(self, add)
			fl.start()

	def drag_put_thumb(self, widget, context, x, y, selection, targetType, eventTime):
		data_type = str(selection.get_data_type())
		if data_type == "text/x-wellpapp-tagguid":
			self.drag_put_thumb_guid(x, y, selection)
		if data_type == "text/x-wellpapp-post-id":
			self.drag_put_thumb_post(selection)

	def drag_put(self, widget, context, x, y, selection, targetType, eventTime, all):
		# @@ check for changing tag values?
		self.apply([((t, None), None) for t in _uni(selection.get_data()).split()], [], all)

	def _drag_get_each(self, model, path, iter, data):
		targetType, l = data
		data = model.get_value(iter, targetType)
		l.append(data)

	def drag_get_list(self, widget, context, selection, targetType, eventTime):
		l = []
		sel = widget.get_selection()
		sel.selected_foreach(self._drag_get_each, (targetType, l))
		# All the examples pass 8, what does it mean?
		selection.set(selection.get_target(), 8, " ".join(l).encode("utf-8"))

	def drag_get_icon(self, widget, context, selection, targetType, eventTime):
		data = ""
		for path in widget.get_selected_items():
			data += self.thumbs[path][0] + " "
		selection.set(selection.get_target(), 8, data[:-1].encode("utf-8"))

	def implications(self, parent, guid):
		dialog = ImplicationsDialog(self, self.client, parent, guid)
		dialog.run()
		self._needs_refresh = dialog.did_something
		dialog.destroy()

	def aliases(self, parent, guid):
		dialog = AliasesDialog(self.client, parent, guid)
		dialog.run()
		dialog.destroy()

	def modify_tag(self, tv, row, *a):
		model = tv.get_model()
		if not model[row][0]:
			return
		pre = prefix(model[row][0])
		guid = clean(model[row][1])[:27]
		tag = self.client.get_tag(guid)
		dialog = TagDialog(self.client, self.window, u"Modify tag", tag.name, tag.type)
		if tag.valuetype:
			vt = markup_escape_text(tag.valuetype)
			mu = "<span color=\"#ff0000\">" + vt + "</span>"
			vtlab = gtk.Label()
			vtlab.set_markup(mu)
			dialog.vbox.pack_start(vtlab, True, True, 0)
		entry = gtk.Entry()
		entry.set_text(tag.name)
		dialog.vbox.pack_start(entry, True, True, 0)
		ag = gtk.AccelGroup()
		dialog.add_accel_group(ag)
		implbutton = button_with_mnemonic(ag, u"_Implications", self.implications, dialog, guid)
		dialog.vbox.pack_start(implbutton, True, True, 0)
		aliasbutton = button_with_mnemonic(ag, u"_Aliases", self.aliases, dialog, guid)
		dialog.vbox.pack_start(aliasbutton, True, True, 0)
		entry.connect("activate", lambda *a: dialog.response(gtk.ResponseType.ACCEPT))
		dialog.show_all()
		self._needs_refresh = False
		if dialog.run() == gtk.ResponseType.ACCEPT:
			new_type = None
			t = dialog.get_tt()
			if t and t != tag.type:
				new_type = t
			new_name = _uni(entry.get_text().strip())
			if new_name == tag.name: new_name = None
			if new_type or new_name:
				try:
					self.client.mod_tag(guid, type=new_type, name=new_name)
					# Don't refresh, just update this tag without re-sorting
					if new_type:
						self.ids[guid].type = new_type
					if new_name:
						self.ids[guid].name = new_name
						model[row][0] = self.fmt_tag(pre + guid)
						model[row][4] = self.txt_tag(pre + guid)
					model[row][3] = self.tag_colour_guid(guid)
				except Exception:
					from traceback import print_exc
					print_exc()
		dialog.destroy()
		if self._needs_refresh: self.refresh()

	def middle_toggle_select(self, widget, event):
		if event.button == 2:
			pathinfo = widget.get_path_at_pos(int(event.x), int(event.y))
			if pathinfo:
				# Why would lists with text and lists with icons have the same interface? Sigh..
				if hasattr(widget, "get_selection"):
					sel = widget.get_selection()
					path = pathinfo[0]
				else:
					sel = widget
					path = pathinfo
				if sel.path_is_selected(path):
					sel.unselect_path(path)
				else:
					sel.select_path(path)
				return True

	def tag_colour(self, type):
		if type in self.type2colour: return self.type2colour[type]
		if isinstance(type, unicode):
			type = type.encode("utf-8")
		return "#%02x%02x%02x" % tuple([int(ord(c) / 1.6) for c in md5(type).digest()[:3]])

	def tag_colour_guid(self, guid):
		return self.tag_colour(self.ids[guid].type)

	def fmt_tag(self, g):
		t = self.ids[clean(g)]
		name = markup_escape_text(prefix(g) + t.name)
		if t.valuetype in (None, "none"): return name
		name = " <span color=\"#ff0000\">★ </span>" + name + " <span color=\"#ff0000\">" + t.valuetype
		if "valuelist" not in t: return name + "</span>"
		if t.localcount == len(t.valuelist) and len(set(t.valuelist)) == 1: # all have the same value
			name += "=" + markup_escape_text(str(t.valuelist[0]))
		else:
			name += " ..."
		return name + "</span>"

	def txt_tag(self, g):
		t = self.ids[clean(g)]
		v = prefix(g) + t.name
		if not t.valuelist: return v
		return " ".join([v + "=" + val.format() for val in set(t.valuelist)])

	def _guid_with_val(self, g):
		t = self.ids[clean(g)]
		if not t.valuelist: return g
		return " ".join([g + "=" + val.format() for val in set(t.valuelist)])

	def put_in_list(self, li):
		lo = getattr(self, "tags_" + li)
		view = getattr(self, "tags_" + li + "view")
		data = []
		for pre, bg in ("", "#ffffff"), ("impl", "#ffd8ee"):
			# unformatted name first for sorting, not put in lo
			data.extend((self.ids[clean(t)].name, self.fmt_tag(t), self._guid_with_val(t), bg, self.tag_colours[clean(t)], self.txt_tag(t), pre) for t in self.taglist[pre + li])
		lo.clear()
		if data:
			view.get_selection().set_mode(gtk.SelectionMode.MULTIPLE)
			# gtk.ListStore doesn't have extend
			for t in sorted(data):
				lo.append(t[1:])
		else:
			# One empty unselectable row to give a bigger drop-target for this group
			view.get_selection().set_mode(gtk.SelectionMode.NONE)
			lo.append()

	def refresh(self):
		self.b_apply.set_sensitive(False)
		PostRefresh(self).start()

	def update_thumb_tooltips(self):
		for thumb in self.thumbs:
			if thumb[0] in self.posts:
				post = self.posts[thumb[0]]
				tip = [post.md5]
				tags = sorted(post.datatags.names.items())
				tip += [name + u": " + unicode(tag.value) for name, tag in tags]
				thumb[2] = u"\n".join(tip)

	def _tagcompute(self, posts, pre):
		def guids(p):
			gs = set(t.pguid for t in p[pre + "fulltags"].values())
			gs.update(t.pguid for t in p[pre + "weaktags"].values())
			return gs
		self.taglist[pre + "all"] = guids(posts[0])
		self.taglist[pre + "any"] = guids(posts[0])
		for p in posts:
			self.taglist[pre + "all"].intersection_update(guids(p))
			self.taglist[pre + "any"].update(guids(p))

	def add_thumbs(self, thumbs):
		for thumb in thumbs:
			self.thumbs.append(thumb)
		self.refresh()

	def known_tag(self, tag):
		return tag["guid"] in self.ids

	def thumb_selected(self, iconview):
		self.update_from_selection()

	def thumb_activated(self, iconview, path):
		if self.fullscreen_open: return
		try:
			self.fullscreen_open = True
			m = self.thumbs[path][0]
			fn = self.client.image_path(m)
			f = FullscreenWindowThread(fn, self)
			f.start()
		except Exception:
			from traceback import print_exc
			print_exc()
			self.fullscreen_open = False

	def update_from_selection(self):
		self._update_from_selection("")
		self._update_from_selection("impl")
		self.put_in_list("allcurrent")
		self.put_in_list("currentother")
		self.put_in_list("other")

	def _update_from_selection(self, pre):
		common = None
		all = set()
		count = 0
		def guids(p):
			gs = set(t.pguid for t in p[pre + "fulltags"].values())
			gs.update(t.pguid for t in p[pre + "weaktags"].values())
			return gs
		for path in self.thumbview.get_selected_items():
			m = self.thumbs[path][0]
			g = guids(self.posts[m])
			if common is None:
				common = g
			else:
				common.intersection_update(g)
			all.update(g)
			count += 1
		if count < 2:
			self.tags_currentotherview.hide()
		else:
			self.tags_currentotherview.show()
		if count == 0:
			self.tags_allcurrentview.hide()
		else:
			self.tags_allcurrentview.show()
		if not common: common = set()
		all.difference_update(self.taglist[pre + "all"])
		all.difference_update(common)
		unique = common.difference(self.taglist[pre + "all"])
		self.taglist[pre + "allcurrent"] = unique
		self.taglist[pre + "currentother"] = all
		other = set(self.taglist[pre + "any"])
		other.difference_update(unique)
		other.difference_update(all)
		other.difference_update(self.taglist[pre + "all"])
		self.taglist[pre + "other"] = other

	def destroy(self, *a):
		gtk.main_quit()

	def fmt_tagalt(self, name, t):
		name = markup_escape_text(name)
		col = self.tag_colour(t)
		return "<span color=\"" + col + "\">" + name + "</span>"

	def show_alts(self, alts):
		self.set_msg(u"")
		mu = " ".join(self.fmt_tagalt(name, t.type) for name, t in alts)
		self.msg.set_markup(mu)

	def tagfield_key(self, tagfield, event):
		return complete_entry(self, self, self.window, tagfield, event)

	def create_tag(self, name):
		dialog = TagDialog(self.client, self.window, u"Create tag", name)
		if dialog.run() == gtk.ResponseType.ACCEPT:
			t = dialog.get_tt()
			if t:
				self.client.add_tag(name, t)
		dialog.destroy()

	def apply_action(self, *a):
		self.set_msg(u"")
		orgtext = _uni(self.tagfield.get_text().strip())
		if not orgtext:
			gtk.main_quit()
			return
		if not len(self.thumbs):
			return
		good = []
		failed = []
		for t in orgtext.split():
			tag = self.client.parse_tag(t)
			if not tag and prefix(t) != u"-":
				self.create_tag(clean(t))
				tag = self.client.parse_tag(t)
			if tag:
				good.append((tag, t))
			else:
				failed.append(t)
		if self.apply(good, failed, False):
			self.tagfield.set_text(orgtext)
		else:
			self.tagfield.set_text(u" ".join(failed))

	def apply(self, good, failed, all):
		if all:
			posts = self.posts
		else:
			posts = [self.thumbs[p][0] for p in self.thumbview.get_selected_items()] or self.posts
		return self._apply(good, failed, posts)

	def _apply(self, good, failed, posts):
		todo = {}
		for tag, t in good:
			try:
				for m in posts:
					if m not in todo: todo[m] = ([], [], [])
					p = prefix(tag[0])
					ctag = clean(tag[0])
					post = self.posts[m]
					if p == "-" and (ctag in post.settags.guids or "~" + ctag in post.settags.guids):
						todo[m][2].append(ctag)
					do_set = False
					if p != "-" and tag[0] not in post.settags.guids:
						do_set = True
					elif tag[1]:
						v = post.tags[tag[0]].value
						if v is None or tag[1].format() != v.format():
							do_set = True
					if do_set:
						if p == "~":
							todo[m][1].append((ctag, tag[1]))
						else:
							assert not p
							todo[m][0].append(tag)
			except Exception:
				from traceback import print_exc
				print_exc()
				failed.append(t)
		bad = False
		todo_m = [m for m, t in todo.items() if t != ([], [], [])]
		if todo_m: self.client.begin_transaction()
		try:
			for m in todo_m:
				full, weak, remove = map(set, todo[m])
				self.client.tag_post(m, full, weak, remove)
		except WellpappError:
			bad = True
		finally:
			if todo_m: self.client.end_transaction()
		self.refresh()
		return bad

	def set_msg(self, msg, bg="#FFFFFF"):
		self.msgbox.modify_bg(gtk.StateType.NORMAL, gdk.color_parse(bg))
		self.msg.set_text(msg)

	def error(self, msg):
		self.set_msg(msg, "#FF4466")

	def progress_begin(self, total_steps):
		self._progress_step = 0
		self._progress_total_steps = total_steps
		self.progress_bar.set_fraction(0.0)
		self.progress_bar.show()
		self.msgbox.hide()

	def progress_step(self):
		if self._progress_step < self._progress_total_steps:
			self._progress_step += 1
			self.progress_bar.set_fraction(self._progress_step / self._progress_total_steps)

	def progress_end(self):
		self._progress_total_steps = 0
		self.progress_bar.hide()
		self.msgbox.show()
		self.b_apply.set_sensitive(True)

	def main(self):
		self.window.show()
		self.tagfield.grab_focus()
		gtk.main()

class TagDialog(gtk.Dialog):
	def __init__(self, client, mainwin, title, tagname, tagtype=None):
		gtk.Dialog.__init__(self, title=title, parent=mainwin, modal=True, destroy_with_parent=True)
		self.add_button(gtk.STOCK_CANCEL, gtk.ResponseType.REJECT)
		self.add_button(gtk.STOCK_OK, gtk.ResponseType.ACCEPT)
		self.set_default_response(gtk.ResponseType.ACCEPT)
		self._client = client
		lab = gtk.Label(label=tagname)
		self.vbox.pack_start(lab, True, True, 0)
		# Workaround for it being impossible to have no selected row, but
		# I still really don't want an immediate enter to be "ok".
		# (Because then when I press enter twice with a misspelled tag
		# I create that tag.)
		self._can_enter = False
		self._cursor_count = 0
		self._tv = self._make_tt_tv(tagtype, tagname)
		self._tv.connect("row-activated", self._accept_maybe)
		self._tv.connect("cursor-changed", self._cursor_changed)
		self.vbox.pack_end(self._tv, True, True, 0)
		self.show_all()

	def _make_tt_tv(self, selname, tagname):
		selpos = None
		ls = gtk.ListStore(TYPE_STRING)
		tt = self._client.metalist(u"tagtypes")
		for pos, t in zip(range(len(tt)), tt):
			ls.append((t,))
			if t == selname: selpos = pos
			if selname is None and tagname[:len(t)] == t: selpos = pos
		tv = gtk.TreeView(model=ls)
		crt = gtk.CellRendererText()
		tv.append_column(gtk.TreeViewColumn(u"Type", crt, text=0))
		if selpos is not None:
			tv.get_selection().select_path((selpos,))
		return tv

	def get_tt(self):
		ls, iter = self._tv.get_selection().get_selected()
		if iter:
			return ls.get_value(iter, 0)

	def _cursor_changed(self, *a):
		# This happens once when the dialog opens, so only set _can_enter the second time
		if self._cursor_count:
			self._can_enter = True
		self._cursor_count += 1

	def _accept_maybe(self, *a):
		if self._can_enter:
			self.response(gtk.ResponseType.ACCEPT)
		else:
			self._can_enter = True

class ImplicationsDialog(gtk.Dialog):
	def __init__(self, tw, client, parent, guid):
		gtk.Dialog.__init__(self, title=u"Implications", parent=parent, modal=True, destroy_with_parent=True)
		self.add_button(gtk.STOCK_CLOSE, gtk.ResponseType.ACCEPT)
		self._tw = tw
		self._client = client
		self.guid = guid
		impl = client.tag_implies(self.guid, True)
		if impl:
			lab = gtk.Label(label=u"Implied by")
			self.vbox.pack_start(lab, True, True, 0)
			rev_impl = gtk.ListStore(TYPE_STRING, TYPE_INT, TYPE_STRING, TYPE_STRING)
			lines = []
			complex = False
			for i in impl[:10]:
				tag = client.get_tag(i.guid, with_prefix=True)
				if i.filter:
					filt = i.filter[0] + unicode(i.filter[1])
				else:
					filt = ""
				val = unicode(i.value or u"")
				if filt or val:
					complex = True
				lines.append((tag.pname, i.prio, filt, val))
			for d in sorted(lines):
				rev_impl.append(d)
			if len(impl) > 10:
				rev_impl.append(("... hidden:", len(impl) - 10, "", ""))
			rev_impl = gtk.TreeView(model=rev_impl)
			if complex:
				headings = ("Tag", "Priority", "Filter", "Value")
			else:
				headings = ("Tag", "Priority")
			for i, n in enumerate(headings):
				crt = gtk.CellRendererText()
				tvc = gtk.TreeViewColumn(n, crt, text=i)
				rev_impl.append_column(tvc)
			rev_impl.get_selection().set_mode(gtk.SelectionMode.NONE)
			self.vbox.pack_start(rev_impl, True, True, 0)
			self.vbox.pack_start(gtk.HSeparator(), True, True, 0)
		lab = gtk.Label(label=u"Implies")
		self.vbox.pack_start(lab, True, True, 0)
		self._ibox = gtk.Table(n_rows=1, n_columns=4)
		self.vbox.pack_start(self._ibox, True, True, 0)
		self._add_name = gtk.Entry()
		self._add_name.connect("activate", self._add)
		self._add_name.connect("key-press-event", self._complete)
		self._add_prio = gtk.Entry()
		self._add_prio.set_width_chars(5)
		self._add_prio.set_text(u"0")
		self._add_prio.connect("activate", self._add)
		self._add_btn = gtk.Button.new_with_label(u"Add")
		self._add_btn.connect("clicked", self._add)
		self._show_impl()
		self.did_something = False
		self.tab_count = 0
		self.tab_alts = []
		self.show_all()

	def _show_impl(self):
		impl = self._client.tag_implies(self.guid)
		[self._ibox.remove(c) for c in self._ibox.get_children()]
		if impl:
			lines = list(map(self._impl_wids, impl))
			self._ibox.resize(len(lines) + 1, 4)
			for row, l in enumerate(sorted(lines)):
				for col, wid in zip(range(4), l[1]):
					self._ibox.attach(wid, col, col + 1, row, row + 1)
			row = len(lines)
		else:
			row = 0
		self._ibox.attach(self._add_name, 0, 1, row, row + 1)
		self._ibox.attach(self._add_prio, 1, 2, row, row + 1)
		self._ibox.attach(self._add_btn, 2, 4, row, row + 1)
		self._ibox.show_all()
		self._add_name.grab_focus()

	def _impl_wids(self, impl):
		name = self._client.get_tag(impl.guid, with_prefix=True).pname
		lab = gtk.Label(label=name)
		if impl.filter or impl.value:
			# Ideally, we'd have support for updating this too.
			prio = gtk.Label(label=str(impl.prio))
			if impl.filter:
				filt = gtk.Label(label=impl.filter[0] + unicode(impl.filter[1]))
			else:
				filt = gtk.Label(label="")
			if impl.value:
				val = gtk.Label(label=unicode(impl.value))
			else:
				val = gtk.Label(label="")
			wids = (lab, prio, filt, val)
		else:
			entry = gtk.Entry()
			entry.set_width_chars(5)
			entry.set_text(unicode(impl.prio))
			entry.connect("activate", lambda *a: self._update(entry, impl.guid))
			update = gtk.Button.new_with_label(u"Update")
			update.connect("clicked", lambda *a: self._update(entry, impl.guid))
			remove = gtk.Button.new_with_label(u"Remove")
			wids = (lab, entry, update, remove)
			remove.connect("clicked", lambda *a: self._remove(wids, impl.guid))
		return name, wids

	def _update(self, entry, guid):
		try:
			prio = int(entry.get_text())
		except ValueError:
			prio = 0
		self._client.add_implies(self.guid, guid, prio)
		self.did_something = True
		entry.set_text(unicode(prio))

	def _remove(self, wids, guid):
		self._client.remove_implies(self.guid, guid)
		self.did_something = True
		for w in wids:
			w.hide()

	def _add(self, *a):
		name = self._add_name.get_text().strip()
		if not name:
			self.emit("close")
			return
		try:
			prio = int(self._add_prio.get_text())
		except ValueError:
			prio = 0
		pre = prefix(name)
		if pre and pre != u"-": return
		tag = self._client.find_tag(name, with_prefix=True)
		if not tag: return
		self._client.add_implies(self.guid, tag, prio)
		self.did_something = True
		self._add_name.set_text(u"")
		self._add_prio.set_text(u"0")
		self._show_impl()

	def show_alts(self, alts):
		# I suppose I should show these, but meh.
		pass

	def _complete(self, tagfield, event):
		return complete_entry(self._tw, self, self, tagfield, event)

class AliasesDialog(gtk.Dialog):
	def __init__(self, client, parent, guid):
		gtk.Dialog.__init__(self, title=u"Aliases", parent=parent, modal=True, destroy_with_parent=True)
		self.add_button(gtk.STOCK_CLOSE, gtk.ResponseType.ACCEPT)
		self._client = client
		self.guid = guid
		self._add_name = gtk.Entry()
		self._add_name.connect("activate", self._add)
		self._add_btn = gtk.Button.new_with_label("Add")
		self._add_btn.connect("clicked", self._add)
		self._list = gtk.Table(n_rows=1, n_columns=2)
		self.vbox.pack_start(self._list, True, True, 0)
		self._refresh()
		self.show_all()

	def _refresh(self):
		tag = self._client.get_tag(self.guid)
		[self._list.remove(c) for c in self._list.get_children()]
		if "alias" in tag:
			lines = [self._wids(n) for n in sorted(tag.alias)]
			self._list.resize(len(lines) + 1, 2)
			for row, wids in zip(range(len(lines)), lines):
				self._list.attach(wids[0], 0, 1, row, row + 1)
				self._list.attach(wids[1], 1, 2, row, row + 1)
			row = len(lines)
		else:
			row = 0
		self._list.attach(self._add_name, 0, 1, row, row + 1)
		self._list.attach(self._add_btn, 1, 2, row, row + 1)
		self._list.show_all()
		self._add_name.grab_focus()

	def _wids(self, name):
		lab = gtk.Label(label=name)
		btn = gtk.Button.new_with_label("Remove")
		btn.connect("clicked", lambda *a: self._rm(name))
		return lab, btn

	def _rm(self, name):
		self._client.remove_alias(name)
		self._refresh()

	def _add(self, *a):
		name = self._add_name.get_text().strip()
		if not name:
			self.emit("close")
			return
		self._client.add_alias(name, self.guid)
		self._add_name.set_text(u"")
		self._refresh()

class TagCompletionDialog(gtk.Dialog):
	def __init__(self, tw, client, parent, word, alts):
		gtk.Dialog.__init__(self, title=u"Completing " + word, parent=parent, modal=True, destroy_with_parent=True)
		self.add_button(gtk.STOCK_CANCEL, gtk.ResponseType.REJECT)
		self.add_button(gtk.STOCK_OK, gtk.ResponseType.ACCEPT)
		self.set_default_response(gtk.ResponseType.ACCEPT)
		self.word = word
		self.alts = gtk.ListStore(*taglisttypes)
		for name, t in sorted(alts):
			m_name = markup_escape_text(name)
			tooltip = [t.name]
			if t.alias:
				tooltip.extend(u"    " + a for a in sorted(t.alias))
			tooltip.append(t.type)
			tooltip.append(u"%d posts, %d weak posts" % (t.posts, t.weak_posts))
			tooltip = [markup_escape_text(l) for l in tooltip]
			for heading, reverse in (("\nImplies", False), ("\nImplied by", True)):
				impl = sorted(client.tag_implies(t.guid, reverse))
				if impl:
					tooltip.append(heading)
					max_show = 12
					if len(impl) > max_show:
						dots = len(impl) - max_show + 1
						impl = impl[:max_show - 1]
					else:
						dots = False
					for i in impl:
						i_t = client.get_tag(i.guid, with_prefix=True)
						iname = markup_escape_text(i_t.pname)
						if i.filter or i.value:
							iname = '<span color="#ff0000">*</span> ' + iname
						tc = tw.tag_colour(i_t.type)
						tooltip.append('    <span color="%s">%s</span>' % (tc, iname))
					if dots:
						tooltip.append("... (%d more)" % (dots,))
			tooltip = "\n".join(tooltip)
			self.alts.append((m_name, t.guid, "#ffffff", tw.tag_colour(t.type), name, tooltip))
		self.altsview = gtk.TreeView(model=self.alts)
		self.altsview.set_tooltip_column(5)
		celltext = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("Alternatives", celltext, markup=0)
		tvc.add_attribute(celltext, "cell-background", 2)
		tvc.add_attribute(celltext, "foreground", 3)
		self.altsview.append_column(tvc)
		self.altsview.connect("row-activated", lambda *a: self.response(gtk.ResponseType.ACCEPT))
		self.scroll = gtk.ScrolledWindow()
		self.scroll.set_policy(gtk.PolicyType.NEVER, gtk.PolicyType.ALWAYS)
		self.scroll.add(self.altsview)
		self.vbox.pack_start(self.scroll, True, True, 0)
		# This can't be the right way. (Why isn't the size automatically right anyway?)
		initial_y = self.get_preferred_size().natural_size.height
		self.show_all() # This changes all reported sizes
		x = max(el.get_preferred_size().natural_size.width for el in (self.altsview, self.scroll, self))
		y = self.altsview.get_preferred_height_for_width(x).natural_height + initial_y + 5
		self.resize(x, min(y, int(client.cfg.tagwindow_height)))

	def get_tt(self):
		ls, iter = self.altsview.get_selection().get_selected()
		if iter:
			return ls.get_value(iter, 4) + u" "
		else:
			return self.word

# This doesn't actually manage the window, it just loads the image for it.
# Only the main thread ever touches visible objects, because I don't want to
# mess with the locking system (thread_enter etc).
class FullscreenWindowThread(Thread):
	def __init__(self, fn, tw):
		Thread.__init__(self)
		self.name = "FullscreenWindow"
		self._fn = fn
		self._tw = tw
		self._win = None

	def run(self):
		loader = None
		try:
			fh = RawWrapper(open(self._fn, "rb"))
			loader = GdkPixbuf.PixbufLoader()
			fh.seek(0, 2)
			l = fh.tell()
			fh.seek(0)
			r = 0
			Z = 1024 * 256
			steps = (l - 1) // Z + 1
			idle_add(self._tw.progress_begin, steps * 2)
			while r < l:
				data = fh.read(Z)
				if not data:
					break
				idle_add(self._tw.progress_step)
				loader.write(data)
				r += len(data)
				idle_add(self._tw.progress_step)
			if r != l:
				msg = "File changed size while reading (%s)" % (self._fn,)
				idle_add(self._tw.error, msg)
				raise RuntimeError(msg)
			pixbuf = loader.get_pixbuf()
			self._win = FullscreenWindow()
			idle_add(self._win._init, self._tw, pixbuf)
			idle_add(self._tw.set_msg, u"")
		except Exception:
			from traceback import print_exc
			print_exc()
			self._cleanup()
		finally:
			if loader:
				loader.close()
			idle_add(self._tw.progress_end)

	def _cleanup(self, *args):
		self._tw.fullscreen_open = False
		if self._win:
			gtk.Window.destroy(self._win, *args)

class FullscreenWindow(gtk.Window):
	def _init(self, tw, pixbuf):
		self._tw = tw
		self.pixbuf = pixbuf
		self.set_events(gdk.EventMask.BUTTON_PRESS_MASK | gdk.EventMask.KEY_PRESS_MASK)
		self.set_title("Fullscreen window")
		self.modify_bg(gtk.StateType.NORMAL, gdk.Color(0, 0, 0))
		self.show()
		self.fullscreen()
		self.set_size_request(1, 1)
		self.pix_w = self.pixbuf.get_width()
		self.pix_h = self.pixbuf.get_height()
		self.image = gtk.Image()
		self.add(self.image)
		self.connect("configure_event", self._on_configure)
		self.connect('key-press-event', self.key_press_event)
		self.connect("button-press-event", self.destroy)
		self.show_all()

	def _on_configure(self, *args):
		self._scale_pixbuf_to_fit_win()

	def _scale_pixbuf_to_fit_win(self):
		win_w, win_h = self.get_size()
		if (win_w < self.pix_w) or (win_h < self.pix_h):
			scalefactor = min(win_w / self.pix_w, win_h / self.pix_h)
			self.scale_pixbuf_to_scalefactor(scalefactor)
		else:
			self.image.set_from_pixbuf(self.pixbuf)
			self.show_all()

	def scale_pixbuf_to_scalefactor(self, scalefactor):
		new_w = int(round(self.pix_w * scalefactor))
		new_h = int(round(self.pix_h * scalefactor))
		new_pixbuf = self.pixbuf.scale_simple(new_w, new_h, GdkPixbuf.InterpType.HYPER)
		self.image.set_from_pixbuf(new_pixbuf)
		self.show_all()

	def key_press_event(self, spin, event):
		key = gdk.keyval_name(event.keyval).lower()
		# All normal keys and a few special
		if len(key) == 1 or key in ("escape", "space", "return"):
			self.destroy()

	def destroy(self, *args):
		self._tw.fullscreen_open = False
		gtk.Window.destroy(self)

class PostRefresh(Thread):
	def __init__(self, tw):
		Thread.__init__(self)
		self.name = "PostRefresh"
		self.daemon = True
		self.client = Client()
		self.tw = tw

	def run(self):
		if not self.tw._progress_total_steps:
			idle_add(self.tw.progress_begin, len(self.tw.thumbs) + 1)
		posts = []
		to_remove = []
		seen = set()
		for ix, t in enumerate(self.tw.thumbs):
			if t[0] in seen:
				to_remove.append(ix)
			else:
				seen.add(t[0])
				p = self.client.get_post(t[0], True)
				if p:
					posts.append(p)
				else:
					to_remove.append(ix)
					idle_add(self.tw.error, u"Post(s) not found")
			idle_add(self.tw.progress_step)
		for ix in reversed(to_remove):
			del self.tw.thumbs[ix]
		if not posts:
			idle_add(self.tw.error, u"No posts found")
			return
		ids = {}
		for t in chain(*[p.tags for p in posts]):
			tt = ids.setdefault(t.guid, t)
			tt.localcount = tt.get("localcount", 0) + 1
			if "value" in t:
				tt.setdefault("valuelist", []).append(t.value)
		self.tw.ids = ids
		self.tw.posts = {p.md5: p for p in posts}
		self.tw.tag_colours = {tg: self.tw.tag_colour_guid(clean(tg)) for tg in ids}
		idle_add(self.tw.progress_step)
		self.tw._tagcompute(posts, "")
		self.tw._tagcompute(posts, "impl")
		idle_add(self.tw.put_in_list, "all")
		idle_add(self.tw.update_from_selection)
		idle_add(self.tw.update_thumb_tooltips)
		idle_add(self.tw.progress_end)

class FileLoaderWorker(Thread):
	def __init__(self, client, tw, q_in, d_out, z):
		Thread.__init__(self)
		self.name = "FileLoaderWorker"
		self.daemon = True
		self._client = client
		self._tw = tw
		self._q_in = q_in
		self._d_out = d_out
		self._z = z

	def run(self):
		try:
			while True:
				d = self._q_in.get(False)
				m = thumb = None
				try:
					m = self._client.postspec2md5(d)
				except Exception as e:
					print(e)
				idle_add(self._tw.progress_step)
				if m:
					try:
						fn = self._client.thumb_path(m, self._z)
						thumb = GdkPixbuf.Pixbuf.new_from_file(fn)
					except Exception as e:
						print(e)
				self._d_out[d] = (m, thumb,)
				self._q_in.task_done()
				idle_add(self._tw.progress_step)
		except Queue.Empty:
			pass

class FileLoader(Thread):
	def __init__(self, tw, argv):
		Thread.__init__(self)
		self.name = "FileLoader"
		self.daemon = True
		self._tw = tw
		self._argv = argv

	def run(self):
		client = Client()
		q_in = Queue.Queue()
		for d in self._argv:
			q_in.put(d)
		d_out = {}
		# 3 per arg: md5, thumb load, post load, plus one extra at the end in tw.refresh
		idle_add(self._tw.progress_begin, len(self._argv) * 3 + 1)
		z = int(client.cfg.thumb_sizes.split()[0])
		for _ in range(min(cpu_count(), len(self._argv))):
			FileLoaderWorker(client, self._tw, q_in, d_out, z).start()
		q_in.join()
		good = True
		ordered_out = [d_out[d] for d in self._argv]
		if any(not m for m, _ in ordered_out):
			idle_add(self._tw.error, u"File(s) not found")
			org_len = len(ordered_out)
			ordered_out = [v for v in ordered_out if v[0]]
			good = False
			for _ in range(org_len - len(ordered_out)):
				idle_add(self._tw.progress_step)
		if not ordered_out:
			idle_add(self._tw.error, u"No files found")
			idle_add(self._tw.progress_end)
			return
		fallback = None
		if any(not tn for _, tn in ordered_out):
			if good:
				idle_add(self._tw.error, u"Thumbs(s) not found")
			good = False
			fallback = gtk.IconTheme().load_icon("image-missing", z, 0)
		thumbs = [(m, tn or fallback, m) for m, tn in ordered_out]
		idle_add(self._tw.add_thumbs, thumbs)
		if good:
			idle_add(self._tw.set_msg, u"")
		else:
			idle_add(self._tw.progress_end)

def main(arg0, argv):
	if len(argv) < 1:
		print("Usage:", arg0, "post-spec [post-spec [...]]")
		return 1
	tw = TagWindow()
	fl = FileLoader(tw, argv)
	fl.start()
	tw.main()



from __future__ import print_function

from wellpapp import Client

def main(arg0, argv):
	if len(argv) < 2:
		print("Usage:", arg0, "post-spec tag [tag [...]]")
		print("or:", arg0, "-r tag post-spec [post-spec [...]]")
		return 1

	def set_tag(full, weak, remove, tag):
		tag = client.parse_tag(tag)
		if not tag: return
		tag, val = tag
		s = full
		if tag[0] == "-":
			s = remove
			tag = tag[1:]
		elif tag[0] == "~":
			s = weak
			tag = tag[1:]
		s.add((tag, val))
		return True

	client = Client()
	full = set()
	weak = set()
	remove = set()
	if argv[0] == "-r":
		if not set_tag(full, weak, remove, argv[1]):
			print("Tag not found")
			return 1
		client.begin_transaction()
		for post in argv[2:]:
			md5 = client.postspec2md5(post)
			if not md5 or not client.get_post(md5):
				print(post, "not found")
				continue
			try:
				client.tag_post(md5, full, weak, remove)
			except Exception:
				print("Failed to set on", post)
		client.end_transaction()
	else:
		md5 = client.postspec2md5(argv[0])
		if not md5 or not client.get_post(md5):
			print("Post not found")
			return 1
		for tag in argv[1:]:
			if not set_tag(full, weak, remove, tag):
				print("Unknown tag", tag)
		if full or weak or remove:
			client.tag_post(md5, full, weak, remove)
